#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai
from __future__ import (unicode_literals, division, absolute_import,
                        print_function)
import six
from six.moves import zip

__license__   = 'GPL v3'
__copyright__ = '2011, Pr.BarnArt based on the Barnes work by Grant Drake'
__docformat__ = 'restructuredtext en'

import time

try:
    # For Python 3.0 and later
    from urllib.request import Request, urlopen
except ImportError:
    # Fall back to Python 2's urllib2
    from six.moves.urllib.request import Request, urlopen

try:
    from urllib.parse import quote,urlencode
except ImportError:
    from six.moves.urllib.parse import quote, urlencode
try:
    from queue import Queue, Empty
except ImportError:
    from six.moves.queue import Queue, Empty  


from lxml.html import fromstring, tostring
from xml.etree import ElementTree

from calibre import as_unicode
from calibre.ebooks.metadata import check_isbn
from calibre.ebooks.metadata.sources.base import Source
from calibre.utils.icu import lower
from calibre.utils.cleantext import clean_ascii_chars


class BOL_DE(Source):

    name                    = 'BOL_DE'
    description             = _('Downloads metadata en covers van BOL.de')
    author                  = 'Pr. BarnArt'
    version                 = (0, 9, 2)
    minimum_calibre_version = (0, 8, 0)

    capabilities = frozenset(['identify', 'cover'])
    touched_fields = frozenset(['title', 'authors', 'identifier:isbn', 'rating','series', 'comments', 'publisher', 'languages','pubdate','tags'])
    has_html_comments = True
    supports_gzip_transfer_encoding = True
    cached_cover_url_is_reliable = True
    
    BASE_URL = 'https://www.bol.de/shop/home/suche/?sq='
    BASE_URL_1 = '&sswg=ANY'
    BASE_ISBN= 'https://www.bol.de/shop/home/suche/?&sq='
    BASE_ISBN_1='&sswg=ANY' 
    BASE_BOL  = 'https://www.bol.de'
    
    def config_widget(self):
        '''
        Overriding the default configuration screen for our own custom configuration
        '''
        from calibre_plugins.BOL_DE.config import ConfigWidget
        return ConfigWidget(self)

    def get_book_url(self, identifiers):
        bol_de_id = identifiers.get('bol_de', None)
        if bol_de_id:
            return ('bol_de', bol_de_id,
                  '%s%s'%(BOL_DE.BASE_BOL, bol_de_id))
                  

    def create_isbn_query(self, log, title=None, authors=None, identifiers={} ):
        isbn = check_isbn(identifiers.get('isbn', None))        
        if isbn is not None:            
            return '%s%s%s'%(BOL_DE.BASE_ISBN,isbn,BOL_DE.BASE_ISBN_1)   
        
    def create_title_query(self, log, title=None, authors=None, identifiers={}):     
        ti = ''
        au = '' 
        print('Titel1:',title)
        if title:          
            # BOL  maybe doesn't cope very well with non ascii names so convert
            #title = get_udc().decode(title)
            print('Titel:',title)
            title_tokens = list(self.get_title_tokens(title,
                                strip_joiners=False, strip_subtitle=True))
            if title_tokens:
                tokens = [quote(t.encode('utf_8') if isinstance(t, six.text_type) else t) for t in title_tokens]               
                ti='+'.join(tokens)
                print('Ti:',ti)
                
        if not ti:
            return None
        if authors:
            # BOL  maybe doesn't cope very well with non ascii names so convert
            log.exception('authors:',authors)
            #authors = [get_udc().decode(a) for a in authors]
            log.exception('authors2:',authors)
            author_tokens = self.get_author_tokens(authors,
                    only_first_author=True)
            if author_tokens:
                
                tokens = [quote(t.encode('utf_8') if isinstance(t, six.text_type) else t) for t in author_tokens]
                au='+'.join(tokens)
                log.exception('au1:',au)
        if not au:
            return None
        return '%s%s%s%s%s'%(BOL_DE.BASE_URL, ti,"+",au,BOL_DE.BASE_URL_1) 

    def get_cached_cover_url(self, identifiers):
        url = None
        bol_de_id = identifiers.get('bol_de', None)
        if bol_de_id is None:
            isbn = identifiers.get('isbn', None)
            if isbn is not None:
               bol_de_id = self.cached_isbn_to_identifier(isbn)
        if bol_de_id is not None:
            url = self.cached_identifier_to_cover_url(bol_de_id)
        return url

    def cached_identifier_to_cover_url(self, id_):
        with self.cache_lock:
            url = self._get_cached_identifier_to_cover_url(id_)
            if not url:
                # Try for a "small" image in the cache
                url = self._get_cached_identifier_to_cover_url('small/'+id_)
            return url

    def _get_cached_identifier_to_cover_url(self, id_):
        # This must only be called once we have the cache lock
        url = self._identifier_to_cover_url_cache.get(id_, None)
        if not url:
            # We could not get a url for this particular B&N id
            # However we might have one for a different isbn for this book
            # Barnes & Noble are not very consistent with their covers and
            # it could be that the particular ISBN we chose does not have
            # a large image but another ISBN we retrieved does.
            key_prefix = id_.rpartition('/')[0]
            for key in self._identifier_to_cover_url_cache.keys():
                if key.startswith('key_prefix'):
                    return self._identifier_to_cover_url_cache[key]
        return url

    def identify(self, log, result_queue, abort, title=None, authors=None,
            identifiers={}, timeout=30):
        '''
        Note this method will retry without identifiers automatically if no
        match is found with identifiers.
        '''
        matches = []
        #need to read the ratings of an indexpage, I cannot trace them on the productpage
        #use ratings for the selected matches, so the order of rating is equal to matches 
        ratings = []
        isbn = check_isbn(identifiers.get('isbn', None))        
        br = self.browser
        query = self.create_isbn_query(log, title=title, authors=authors,
                identifiers=identifiers)
        isbn_match_failed = True
        if query is not None and isbn is not None: 
            log.info('Querying: %s'%query)
            response = br.open_novisit(query, timeout=timeout)
            try:    
                raw = response.read().strip()                               
                raw = raw.decode('windows-1252', errors='replace')                  
                if not raw:
                    log.error('Failed to get raw result for query: %r'%query)
                    return
                root = fromstring(clean_ascii_chars(raw))
                #checknode_node=root.xpath('//input [@class="b8_text" and @name="st"]')
                
                url_node=root.xpath('//suchergebnis-liste/ul/li/ul/li[1]/a/@href')
                #url_node=root.xpath('//div[@class="productList socs_product_list" and @impression="product"]/div/div/a/h2/span[@class="oProductTitle"]/../../@href')
                print(('url_node:',url_node))
                if url_node:
                    isbn_url= self.BASE_BOL+url_node[0]
                    print(('isbn_url:', isbn_url))
                else:
                    isbn_url=query;
                rating = 0
                ratings.append(rating)
                matches.append(isbn_url)   
                isbn_match_failed = False
            except:
                msg = 'Failed to parse BOL.com page for query: %r'%query
                log.exception(msg)
                return msg                  
    
            # For successful ISBN based searches we have already done everything we need to
            # So anything from this point below is for title/author based searches.
        if  isbn_match_failed:
            query = self.create_title_query(log, title=title, authors=authors,
                identifiers=identifiers )
            if query is None:
                log.error('Insufficient metadata to construct titlequery')
                return
            log.info('Querying title based: %s'%query)
            response = br.open_novisit(query, timeout=timeout)
            try:                    
                raw = response.read().strip()
                raw = raw.decode('windows-1252', errors='replace')
                if not raw:
                    log.error('Failed to get raw result for query: %r'%query)
                    return
                root = fromstring(clean_ascii_chars(raw))
            except:
                msg = 'Failed to parse BOL.com page for titlebased query: %r'%query
                log.exception(msg)
                return msg
            # Now grab the matches from the search results, provided the
            # title and authors appear to be for the same book
            self._parse_search_results(log, title, authors, root, matches,ratings,timeout,query)

        if abort.is_set():
            return
        if not matches:        
            if identifiers and title and authors:
                log.info('No matches found with identifiers, retrying using only'
                        ' title and authors')
                return self.identify(log, result_queue, abort, title=title,
                        authors=authors, timeout=timeout)
            log.error('No matches found with query: %r'%query)
            return

        from calibre_plugins.BOL_DE.worker import Worker
        #need to mix url and ratings for the worker class
        combos= list(zip(matches,ratings))
        workers = [Worker(combo, result_queue, br, log, i, self) for i, combo in
                enumerate(combos)]

        for w in workers:
            w.start()
            # Don't send all requests at the same time
            time.sleep(0.1)

        while not abort.is_set():
            a_worker_is_alive = False
            for w in workers:
                w.join(0.2)
                if abort.is_set():
                    break
                if w.is_alive():
                    a_worker_is_alive = True
            if not a_worker_is_alive:
                break

        return None

    def _parse_search_results(self, log, orig_title, orig_authors, root, matches, ratings, timeout,query):
        def ismatch(title):
             #only if the title exact matches
            match = False
            if lower(title)==lower(orig_title):
               match= True
            return match 
        def iscat(cattxt):
            cat = False
            txt= lower(cattxt)
            if txt.find('epub')>=0:
                cat=True
            if txt.find('gebundene')>=0:
                cat=True
            if txt.find('book')>=0:
                cat=True
            if txt.find('einband')>=0:
                cat=True
            if txt.find('buch')>=0:
                cat=True
            if txt.find('rbuch')>=0:
                cat=False
            return cat
        import calibre_plugins.BOL_DE.config as cfg
        max_results = cfg.plugin_prefs[cfg.STORE_NAME][cfg.KEY_MAX_DOWNLOADS]
        results = root.xpath('//suchergebnis-liste/ul/li') 
        rating = 0
        print('results:',results)
        if results:
            for result in results:
                print('result1:',result)
                #print(ElementTree.tostring(result, method='xml'))
                title_node= result.xpath('section[@class="artikel-details"]/h3')
                print ('title_node:',title_node)
                title = title_node[0].text_content()
                print('titel:',title)
                pos=title.find(': ')
                if pos>0:
                    title= title[pos+2:]
                print('titel3:',title)
                pos=title.find('(')
                if pos>0:
                    title=title[:pos]
                pos=title.find('/')
                if pos>0:
                    title=title[:pos]
                  
                print('titel4:',title)   
                title=title.replace('\n','').strip()
                print('titel5:',title)  
                if ismatch(title):
                    print ('match')
                    stringtxt='ul [@class="weitere-formate no-bullets"]/li/a' 
                    format_node= result.xpath(stringtxt)
                    print('format_node:',format_node)
                    if format_node:
                        format=format_node[0].text_content().strip().lower()
                        print('format:',format)
                        if iscat(format):
                            print('iscat')
                            
                            rating_node=result.xpath('section[@class="artikel-details"]/p [@class="element-rating-standard"]/@rating')
                            print('rating_node:',rating_node)
                            if  rating_node:
                                rating=rating_node[0]
                                print('rating:',rating)
                            url_node=result.xpath('ul [@class="weitere-formate no-bullets"]/li/a/@href')
                            print ('url_node:',url_node)
                            if url_node:
                                print ('url_node:',url_node)
                                urlquery= '%s%s'%(self.BASE_BOL,url_node[0])
                                print('urlquery:',urlquery)
                                matches.append(urlquery)   
                                ratings.append(rating)
                                if len(matches) >= max_results:
                                    return

    def download_cover(self, log, result_queue, abort,
            title=None, authors=None, identifiers={}, timeout=30):
        cached_url = self.get_cached_cover_url(identifiers)
        if cached_url is None:
            log.info('No cached cover found, running identify')
            rq = Queue()
            self.identify(log, rq, abort, title=title, authors=authors,
                    identifiers=identifiers)
            if abort.is_set():
                return
            results = []
            while True:
                try:
                    results.append(rq.get_nowait())
                except Empty:
                    break
            results.sort(key=self.identify_results_keygen(
                title=title, authors=authors, identifiers=identifiers))
            for mi in results:
                cached_url = self.get_cached_cover_url(mi.identifiers)
                if cached_url is not None:
                    break
        if cached_url is None:
            log.info('No cover found')
            return

        if abort.is_set():
            return
        br = self.browser
        log('Downloading cover from:', cached_url)
        try:
            cdata = br.open_novisit(cached_url, timeout=timeout).read()
            result_queue.put((self, cdata))
        except:
            log.exception('Failed to download cover from:', cached_url)

if __name__ == '__main__': # tests
    # To run these test use:
    # calibre-debug -e __init__.py
    from calibre.ebooks.metadata.sources.test import (test_identify_plugin,
            title_test, authors_test, series_test)
    test_identify_plugin(BOL_DE.name,
        [
            ( # A book with no ISBN specified
                {'title':"Die Befreiung von Hibernia", 'authors':['John Flanagan']},
                [title_test("Die Befreiung von Hibernia", exact=True), 
                 authors_test(['John Flanagan']),
                 series_test('',None)
                ]
            ),
            ( # A book with no ISBN specified
                {'title':"Der Mann, der lächelte", 'authors':['Henning Mankell']},
                [title_test("Tea-Bag", exact=True), 
                 authors_test(['Henning Mankell'])
                ]
            ),
            
            ( # A book with ISBN specified
            
                {'identifiers':{'isbn': '9783843702171'},
                'title':"Knast", 'authors':['Joe Bausch']},
                [title_test("Knast", exact=True), 
                 authors_test(['Joe Bausch']),
                 series_test('',None)
                ]
            ),
            ( # A book with no ISBN specified         
                
                {'title':"Knast", 'authors':['Joe Bausch']},
                [title_test("Knast", exact=True), 
                 authors_test(['Joe Bausch']),
                 series_test('',None)
                ]
            ),
            ( # A book with no ISBN specified
                {'title':"Harry Potter und die Kammer des Schreckens", 'authors':['Joanne K. Rowling']},
                [title_test("Tea-Bag", exact=True), 
                 authors_test(['Henning Mankell'])
                ]
            ),
            ( # A book with no ISBN specified
                {'title':"Lolek und Bolek - Am Orinoko", 'authors':['Leszek Mech']},
                [title_test("Tea-Bag", exact=True), 
                 authors_test(['Henning Mankell'])
                ]
            ),
            ( # A book with an ISBN
                {'identifiers':{'isbn': '9783874079860'},
                    'title':'Endlich fit', 'authors':['Juergen Seibold']},
                [title_test('Der augensammler',
                    exact=True), authors_test(['Sebastian Fitzek']),
                    series_test('Harry Potter', 2.0)]
            ),
            
             ( # A book with no ISBN specified
                {'title':"Harry Potter 2 und die Kammer des Schreckens", 'authors':['Joanne K. Rowling']},
                [title_test("Tea-Bag", exact=True), 
                 authors_test(['Henning Mankell'])
                ]
            ),
            ( # A book with no ISBN specified
                {'title':"Endlich fit", 'authors':['Jürgen Seibold']},
                [title_test("Tea-Bag", exact=True), 
                 authors_test(['Henning Mankell'])
                ]
            ),
            ( # A book with no ISBN specified
                {'title':"Knast", 'authors':['Joe Bausch']},
                [title_test("Knast", exact=True), 
                 authors_test(['Joe Bausch'])
                ]
            )
        ])


